#include "distdata.h"
#include "memutils.h"
#include "errutils.h"
#include "ioutils.h"

#include <stdio.h>
#include <string.h>

/*
 * Converts direct acces index set to sequential index set.
 * O:  set .. sequential index set
 * I: lset .. length of 'set'
 *      pl .. direct access index set (place list)
 *     lpl .. length of 'pl'
 * e.g.: pl = {0, -1, -1, 1, -1, 3, -1, 2, -1}
 *             0   1   2  3   4  5   6  7   8
 *      -> set = {0, 3, 7, 5}
 */
int32 dd_placeList2Set( int32 *set, int32 lset, int32 *pl, int32 lpl )
{
  int32 i;

  for (i = 0; i < lpl; i++) {
    if (pl[i] >= 0) {
      set[pl[i]] = i;
#ifdef DEBUG_DD
      output( "%d %d %d", i, pl[i], set[pl[i]] );
#endif
    }
  }

  return( 0 );
}

/*
 * Converts sequential index set to direct acces index set.
 * O:   pl .. direct access index set (place list)
 * I:  lpl .. length of 'pl'
 *     set .. sequential index set
 *    lset .. length of 'set'
 */
int32 dd_set2PlaceList( int32 *pl, int32 lpl, int32 *set, int32 lset )
{
  int32 i;

  for (i = 0; i < lpl; i++) {
    pl[i] = -1;
  }
  for (i = 0; i < lset; i++) {
    pl[set[i]] = i;
  }

  return( 0 );
}

/*
 * Compute local vector size for each thread.
 * O:       size .. local vector sizes
 * I: sizeGlobal .. global vector size
 *       nThread .. number of threads
 */
int32 dd_localSize( int32 *size, int32 sizeGlobal, int32 nThread )
{
  int32 i, is;
  float ds, s;

  ds = (float) sizeGlobal / nThread;

  s = ds;
  is = 0;
  for (i = 0; i < nThread; i++) {
    is += size[i] = (int32) s - is;
    s += ds;
  }
  if ((sizeGlobal - is) != 0) {
    size[nThread-1] += sizeGlobal - is;
  }
  for (i = 0; i < nThread; i++) {
#ifdef DEBUG_DD
    output( "%d", size[i] );
#endif
  }

  return( 0 );
}

/*
 * Compute local vector indices for each task.
 * O:     set .. local index sets
 * I:    size .. number of local indices of each task
 *    nThread .. number of threads
 */
int32 dd_localIndexSet( int32 **set, int32 *size, int32 nThread )
{
  int32 i, j;
  int32 k;

  k = 0;
  for (i = 0; i < nThread; i++) {
    for (j = 0; j < size[i]; j++) {
      set[i][j] = k;
      k++;
    }
  }

  return( 0 );
}

/*
 * Create local communication index set (left-right - it contains column
 * indices of LR interface matrix, i.e. col. ind. of nonzeros in global matrix
 * in all local rows) for matrix-vector products.
 * O: setLR .. local communication index set (place list form)
 *      nnz .. mumber of local comm. indices
 * I:     m .. global matrix
 *      set .. local index set
 *     size .. size of 'set'
 * !!! 'setLR' is allocated here.
 */
int32 dd_mCSR_localCommLRIndexSet( int32 *setLR, int32 *nnz,
				   Matrix *m, int32 *set,
				   int32 size )
{
  StorageCSR *data = (StorageCSR *) m->data;
  int32 i, j;
  int32 *pc;

  memset( setLR, 0, m->nCol * sizeof( int32 ) );
  for (i = 0; i < size; i++) {
    pc = &data->col[data->row[set[i]]];
    for (j = 0; j < data->nInRow[set[i]]; j++) {
      setLR[pc[j]] = 1;
    }
  }
  for (i = 0; i < size; i++) {
    setLR[set[i]] = 0;
  }
  j = 0;
  for (i = 0; i < m->nCol; i++) {
    if (setLR[i]) {
      setLR[i] = j;
      j++;
    } else {
      setLR[i] = -1;
    }
  }

  *nnz = j;

  return( 0 );
}

/*
 * Create communication list for vector refresh for matrix-vector products.
 * Can be used for both input and output lists.
 * O:     pcl .. pointer to communication list
 * I:    rank .. rank of this thread
 *    nThread .. number of threads
 *     sizePL .. size of a place list
 *      sizeI .. size of sender index set
 *       setI .. sender index set
 *      sizeO .. sizes of index sets of each receiver
 *       setO .. index sets of each receiver
 * !!! 'pcl' is allocated here.
 */
int32 dd_createCommList( CommList **pcl, int32 rank, int32 nThread,
			 int32 sizePL,
			 int32 sizeI, int32 *setI,
			 int32 *sizeO, int32 **setO )
{
  CommList *cl;
  int32 *pl;
  int32 *tid;
  int32 nn;
  int32 i, j, k, ii, jj;

  cl = *pcl = (CommList *) allocMem( sizeof( CommList ) );
  if (cl == 0) {
    g_error = ErrorAlloc;
    return( -1 );
  }

  pl = (int32 *) allocMem( sizePL * sizeof( int32 ) );

  /* Convert sender index set to place list. */
  dd_set2PlaceList( pl, sizePL, setI, sizeI );

  tid = (int32 *) allocMem( nThread * sizeof( int32 ) );

  /*
   * Number of communication neighbours.
   */
  memset( tid, 0, nThread * sizeof( int32 ) );
  for (i = 0; i < nThread; i++) {
    if (i == rank) continue;
    for (j = 0; j < sizeO[i]; j++) {
      if (pl[setO[i][j]] >= 0) {
	tid[i] = 1;
	break;
      }
    }
  }
  nn = 0;
  for (i = 0; i < nThread; i++) {
    nn += tid[i];
  }

  cl->nList = nn;
  cl->cs = (CommSet *) allocMem( nn * sizeof( CommSet ) );

  for (i = 0; i < nn; i++) {
    cl->cs[i].size = 0;
  }

  /*
   * Sizes of communication index sets.
   */
  k = -1;
  for (i = 0; i < nThread; i++) {
    if (i == rank) continue;
    if (tid[i]) {
      k++;
      cl->cs[k].rank = i;
    }
    for (j = 0; j < sizeO[i]; j++) {
      if (pl[setO[i][j]] >= 0) {
	cl->cs[k].size++;
      }
    }
  }

  for (i = 0; i < nn; i++) {
    cl->cs[i].set = (int32 *) allocMem( cl->cs[i].size * sizeof( int32 ) );
  }

  /*
   * Communication index sets.
   */
  k = -1;
  for (i = 0; i < nThread; i++) {
    if (i == rank) continue;
    if (tid[i]) {
      k++;
    }
    ii = 0;
    for (j = 0; j < sizeO[i]; j++) {
      jj = setO[i][j];
      if (pl[jj] >= 0) {
	cl->cs[k].set[ii] = jj;
	ii++;
      }
    }
  }

  freeMem( pl );
  freeMem( tid );

  return( 0 );
}

/*
 * Distribute global matrix.
 * O:     pmLocal .. pointer to local matrix
 *           pmLR .. pointer to local LR interface matrix
 *           pclI .. pointer to input communication list
 *           pclO .. pointer to output communication list
 *     psizeLocal .. pointer to local size
 *      psetLocal .. pointer to local set
 *        psizeLR .. pointer to local LR size
 *         psetLR .. pointer to local LR set
 *    psizeGlobal .. pointer to global vector size
 * I:           m .. global batrix
 *        nThread .. number of threads
 * !!! 'pnLocal', 'pmLR', 'pclI', 'pclO',
 * !!! 'psetLocal', 'psetLR' are allocated here.
 * e.g.:
 *      |       |   |                   | |       |
 *      |       |   |                   | | v_i^l |
 *      | w_i   | = |  A_i^l A_i A_i^r  | | v_i   |
 *      |       |   |                   | | v_i^r |
 *      |       |   |                   | |       |
 * pmLocal <- A_i, pmLR <- A_i^l + A_i^r
 */
int32 dd_distribMatrix( Matrix **pmLocal, Matrix **pmLR,
			CommList **pclI, CommList **pclO,
			int32 *psizeLocal, int32 **psetLocal,
			int32 *psizeLR, int32 **psetLR,
			int32 *psizeGlobal, Matrix *m, int32 nThread )
{
  Matrix *mTmp;
  int32 *sizeI;
  int32 **setI;
  int32 *sizeO;
  int32 **setO;
  int32 *setLR;
  int32 me;
  int32 i, j, k, sizePL;
  CommList *cl;
  int32 nIncoming, *expectFrom;
  int32 ret;

  /*
   * All threads.
   */
  output( "Distribute matrix." );

  ret = itc_threadRank( &me );
  if (ret) return( ret );

  setI = (int32 **) allocMem( nThread * sizeof( int32 * ) );
  setO = (int32 **) allocMem( nThread * sizeof( int32 * ) );
  sizeI = (int32 *) allocMem( nThread * sizeof( int32 ) );
  sizeO = (int32 *) allocMem( nThread * sizeof( int32 ) );

  if (me == 0) {
    /*
     * Master.
     */

    /* Compute local sizes for each thread. */
    dd_localSize( sizeI, m->nCol, nThread );

    for (i = 0; i < nThread; i++) {
      setI[i] = (int32 *) allocMem( sizeI[i] * sizeof( int32 ) );
    }
    /* Create local index set for each thread. */
    dd_localIndexSet( setI, sizeI, nThread );

    setLR = (int32 *) allocMem( m->nCol * sizeof( int32 ) );

    /*
     * Create local (LR) matrices, local LR sets for each task.
     * Send matrices to each slave.
     */
    for (i = 0; i < nThread; i++) {
      dd_mCSR_localCommLRIndexSet( setLR, &sizeO[i], m, setI[i],
				   sizeI[i] );

      setO[i] = (int32 *) allocMem( sizeO[i] * sizeof( int32 ) );
      dd_placeList2Set( setO[i], sizeO[i], setLR, m->nCol );

      ret = mCSR_submatrix( &mTmp, m, sizeI[i], sizeI[i],
			    setI[i], setI[i], 0 );
      if (ret) return( ret );

      if (i != me) {
	ret = itc_mCSR_sendMatrix( mTmp, i, ITC_TagMatrix );
	if (ret) return( ret );
	mCSR_freeDestroy( &mTmp );
      } else {
	*pmLocal = mTmp;
#ifdef DEBUG_DD
	mCSR_save( mTmp, "M_m.bak" );
#endif
      }

      ret = mCSR_submatrix( &mTmp, m, sizeI[i], sizeO[i],
			    setI[i], setLR, 1 );
      if (ret) return( ret );

      if (i != me) {
	ret = itc_mCSR_sendMatrix( mTmp, i, ITC_TagMatrix + 10 );
	if (ret) return( ret );	
	mCSR_freeDestroy( &mTmp );
      } else {
	*pmLR = mTmp;
#ifdef DEBUG_DD
	mCSR_save( mTmp, "Mlr_m.bak" );
#endif
      }
    }

#ifdef DEBUG_DD
    for (i = 0; i < nThread; i++) {
      for (j = 0; j < sizeI[i]; j++) {
	output( "setI: %d %d %d", i, j, setI[i][j] );
      }
      for (j = 0; j < sizeO[i]; j++) {
	output( "setO: %d %d %d", i, j, setO[i][j] );
      }
    }
#endif

    /*
     * Send all sets to each slave.
     * (Local sets of others are not needed now, but later who knows.)
     */
    for (i = 1; i < nThread; i++) {
      for (j = 0; j < nThread; j++) {
	ret = itc_sendArray( setI[j], sizeI[j], ITC_TypeInt32,
			     i, ITC_TagArray+2*(2*i*nThread+j) );
	if (ret) return( ret );
      }
      for (j = 0; j < nThread; j++) {
	ret = itc_sendArray( setO[j], sizeO[j], ITC_TypeInt32,
			     i, ITC_TagArray+2*((2*i+1)*nThread+j) );
	if (ret) return( ret );
      }
    }

    /*
     * Send global size to each slave.
     */
    sizePL = m->nCol;
    for (i = 1; i < nThread; i++) {
      ret = itc_sendValue( &sizePL, ITC_TypeInt32, i, ITC_TagValue );
      if (ret) return( ret );
    }

  } else {
    /*
     * Slaves.
     */

    /*
     * Receive local and LR matrix.
     */
    ret = itc_mCSR_recvMatrix( pmLocal, 0, ITC_TagMatrix );
    if (ret) return( ret );
#ifdef DEBUG_DD
    mCSR_save( *pmLocal, "M_s.bak" );
#endif

    ret = itc_mCSR_recvMatrix( pmLR, 0, ITC_TagMatrix + 10 );
    if (ret) return( ret );
#ifdef DEBUG_DD
    mCSR_save( *pmLR, "Mlr_s.bak" );
#endif

    /*
     * Receive all sets.
     */
    for (j = 0; j < nThread; j++) {
      ret = itc_recvArray( (void **) &setI[j], &sizeI[j], ITC_TypeInt32,
			   0, ITC_TagArray+2*(2*me*nThread+j) );
      if (ret) return( ret );
    }
    for (j = 0; j < nThread; j++) {
      ret = itc_recvArray( (void **) &setO[j], &sizeO[j], ITC_TypeInt32,
			   0, ITC_TagArray+2*((2*me+1)*nThread+j) );
      if (ret) return( ret );
    }

#ifdef DEBUG_DD
    for (i = 0; i < nThread; i++) {
      for (j = 0; j < sizeI[i]; j++) {
	output( "setI: %d %d %d", i, j, setI[i][j] );
      }
      for (j = 0; j < sizeO[i]; j++) {
	output( "setO: %d %d %d", i, j, setO[i][j] );
      }
    }
#endif

    /*
     * Receive global size.
     */
    ret = itc_recvValue( &sizePL, ITC_TypeInt32, 0, ITC_TagValue );
    if (ret) return( ret );

  }

#ifdef DEBUG_DD
  output( "%d", sizePL );
#endif

  output( "1. phase done." );

  /*
   * All threads.
   */
  *psizeGlobal = sizePL;

  /* Create output communication list */
  ret = dd_createCommList( pclO, me, nThread, sizePL,
			   sizeI[me], setI[me], sizeO, setO );
  cl = *pclO;

#ifdef DEBUG_DD
  output( "CommList out:" );
  for (i = 0; i < cl->nList; i++) {
    output( "%d %d", cl->cs[i].rank, cl->cs[i].size );
    for (j = 0; j < cl->cs[i].size; j++) {
      output( "%d", cl->cs[i].set[j] );
    }
  }
#endif

  /*
   * Assign local sets to output params and free others'.
   */
  for (i = 0; i < nThread; i++) {
    if (i == me) {
      *psizeLocal = sizeI[i];
      *psetLocal = setI[i];
    } else freeMem( setI[i] );
  }
  freeMem( setI );
  freeMem( sizeI );
   
  for (i = 0; i < nThread; i++) {
    if (i == me) {
      *psizeLR = sizeO[i];
      *psetLR = setO[i];
    } else freeMem( setO[i] );
  }
  freeMem( setO );
  freeMem( sizeO );

  /*
   * Big set swapping - output set of one task is input set of another,
   * so send it to him...
   * Preparation phase.
   */
  if (me == 0) {
    /*
     * Master.
     */
    int32 *ncnAll, **cnInfoAll, *nt;

    /*
     * Receive ranks of somm. output targets of each thread, count them and
     * tell each thread how many comm. lists it should expect and
     * who would be sender.
     */
    nt = (int32 *) allocMem( nThread * sizeof( int32 ) );
    ncnAll = (int32 *) allocMem( nThread * sizeof( int32 ) );
    cnInfoAll = (int32 **) allocMem( nThread * sizeof( int32 * ) );
    for (i = 1; i < nThread; i++) {
      ret = itc_recvArray( (void **) &cnInfoAll[i], &ncnAll[i], ITC_TypeInt32,
			   i, ITC_TagArray );
      if (ret) return( ret );
    }
    ncnAll[0] = cl->nList;
    cnInfoAll[0] = (int32 *) allocMem( cl->nList * sizeof( int32 ) );
    for (i = 0; i < cl->nList; i++) {
      cnInfoAll[0][i] = cl->cs[i].rank;
    }
    memset( nt, 0, nThread * sizeof( int32 ) );
    for (i = 0; i < nThread; i++) {
      for (j = 0; j < ncnAll[i]; j++) {
	nt[cnInfoAll[i][j]]++;
      }
    }
    nIncoming = nt[0];
    for (i = 1; i < nThread; i++) {
      ret = itc_sendValue( &nt[i], ITC_TypeInt32, i, ITC_TagValue );
      if (ret) return( ret );
    }
    expectFrom = (int32 *) allocMem( nIncoming * sizeof( int32 ) );
    k = 0;
    for (i = 0; i < nThread; i++) {
      for (j = 0; j < ncnAll[i]; j++) {
	if (cnInfoAll[i][j] == 0) {
	  expectFrom[k] = i;
	  k++;
	} else {
	  ret = itc_sendValue( &i, ITC_TypeInt32, cnInfoAll[i][j],
			       ITC_TagValue+1 );
	  if (ret) return( ret );
	}
      }
    }
  } else {
    /*
     * Slaves.
     */
    int32 *cnInfo;

    /*
     * Tell master ranks of all output targets.
     */
    cnInfo = (int32 *) allocMem( cl->nList * sizeof( int32 ) );
    for (i = 0; i < cl->nList; i++) {
      cnInfo[i] = cl->cs[i].rank;
    }
    ret = itc_sendArray( cnInfo, cl->nList, ITC_TypeInt32,
			 0, ITC_TagArray );
    if (ret) return( ret );
    freeMem( cnInfo );

    ret = itc_recvValue( &nIncoming, ITC_TypeInt32, 0, ITC_TagValue );
    if (ret) return( ret );

    /*
     * Who will send me input lists?
     */
    expectFrom = (int32 *) allocMem( nIncoming * sizeof( int32 ) );
    for (i = 0; i < nIncoming; i++) {
      ret = itc_recvValue( &expectFrom[i], ITC_TypeInt32, 0, ITC_TagValue+1 );
      if (ret) return( ret );
    }
  }
  /*
   * All threads.
   */

  /*
   * Big set swapping - output set of one task is input set of another,
   * so send it to him...
   * Swapping phase - receive all input comm. lists.
   */
  /*  itc_barrier(); */
#ifdef DEBUG_DD
  output( "incoming: %d", nIncoming );
  for (i = 0; i < nIncoming; i++) {
    output( "expect from: %d", expectFrom[i] );
  }
#endif
  for (i = 0; i < cl->nList; i++) {
    ret = itc_sendValue( &me, ITC_TypeInt32, cl->cs[i].rank, ITC_TagValue+2 );
    if (ret) return( ret );
#ifdef DEBUG_DD
    output( "out: cl->cs[i].rank: %d", cl->cs[i].rank );
#endif
    ret = itc_sendArray( cl->cs[i].set, cl->cs[i].size, ITC_TypeInt32,
			 cl->cs[i].rank, ITC_TagArray );
    if (ret) return( ret );
  }

  cl = *pclI = (CommList *) allocMem( sizeof( CommList ) );
  cl->nList = nIncoming;
  cl->cs = (CommSet *) allocMem( nIncoming * sizeof( CommSet ) );

  for (i = 0; i < nIncoming; i++) {
    ret = itc_recvValue( &cl->cs[i].rank, ITC_TypeInt32,
			 expectFrom[i], ITC_TagValue+2 );
    if (ret) return( ret );
#ifdef DEBUG_DD
    output( "in: cl->cs[i].rank: %d", cl->cs[i].rank );
#endif
    ret = itc_recvArray( (void **) &cl->cs[i].set, &cl->cs[i].size,
			 ITC_TypeInt32,
			 cl->cs[i].rank, ITC_TagArray );
    if (ret) return( ret );
  }

  freeMem( expectFrom );

#ifdef DEBUG_DD
  output( "CommList in:" );
  for (i = 0; i < cl->nList; i++) {
    output( "%d %d", cl->cs[i].rank, cl->cs[i].size );
    for (j = 0; j < cl->cs[i].size; j++) {
      output( "%d", cl->cs[i].set[j] );
    }
  }
#endif
  /*  itc_barrier(); */

  output( "2. phase done." );
  return( ret );
}

/*
 * Translate global indices in communication list to local.
 * O:        clL .. communication list with local indices
 * I:        clG .. communication list with global indices
 *          size .. size of 'set'
 *           set .. local index set
 *    sizeGlobal .. global size
 */
int32 dd_translateGlobal2Local( CommList *clL, CommList *clG,
				int32 size, int32 *set, int32 sizeGlobal )
{
  int32 *pl;
  int32 i, j;

  if (clL->nList != clG->nList) {
    g_error = ErrorBadLimit;
    return( -1 );
  }

  pl = (int32 *) allocMem( sizeGlobal * sizeof( int32 ) );

  dd_set2PlaceList( pl, sizeGlobal, set, size );

  for (i = 0; i < clL->nList; i++) {
    for (j = 0; j < clL->cs[i].size; j++) {
      clL->cs[i].set[j] = pl[clG->cs[i].set[j]];
    }
  }

  freeMem( pl );

  return( 0 );
}

/*
 * Distribute global vector.
 * O:        pv .. pointer to local vector
 * I:        vG .. global vector
 *    sizeLocal .. size of 'setLocal'
 *     setLocal .. local index set
 *      nThread .. number of threads
 * !!! 'pv' is allocated here.
 */
int32 dd_distribVector( Vector **pv, Vector *vG,
			int32 sizeLocal, int32 *setLocal, int32 nThread )
{
  int32 i, j;
  Vector *v;
  float64 *bufF;
  int32 *bufI;
  int32 size;
  int32 me;
  int32 ret;

  /*
   * All threads.
   */
  output( "Distribute vector." );

  ret = itc_threadRank( &me );
  if (ret) return( ret );

#ifdef DEBUG_DD
  output( "%d", sizeLocal );
#endif

  ret = v_createAlloc( pv, sizeLocal );
  if (ret) return( ret );
  v = *pv;

  if (me == 0) {
    /*
     * Master.
     */
    bufF = (float64 *) allocMem( vG->nItem * sizeof( float64 ) );
    bufI = (int32 *) allocMem( vG->nItem * sizeof( int32 ) );
    for (i = 1; i < nThread; i++) {
      ret = itc_recvArrayALNA( bufI, &size, ITC_TypeInt32,
			       i, ITC_TagArray );
      if (ret) return( ret );
      for (j = 0; j < size; j++) {
	bufF[j] = vG->val[bufI[j]];
      }
      ret = itc_sendArray( bufF, size, ITC_TypeFloat64,
			   i, ITC_TagArray );
      if (ret) return( ret );
    }
    freeMem( bufI );
    freeMem( bufF );

    for (i = 0; i < sizeLocal; i++) {
      v->val[i] = vG->val[setLocal[i]];
    }

#ifdef DEBUG_DD
    v_save( v, "B_m.bak" );
#endif

  } else {
    /*
     * Slaves.
     */
    ret = itc_sendArray( setLocal, sizeLocal, ITC_TypeInt32,
			 0, ITC_TagArray );
    if (ret) return( ret );
    ret = itc_recvArrayNA( v->val, &v->nItem, v->nAlloc, ITC_TypeFloat64,
			 0, ITC_TagArray );
    if (ret) return( ret );

#ifdef DEBUG_DD
    v_save( v, "B_s.bak" );
#endif

  }

  return( 0 );
}

/*
 * Collect global vector.
 * O: pvG .. pointer to global vector
 * I:   v .. local vector
 *    sizeGlobal .. global size
 *      setLocal .. local index set
 *       nThread .. number of threads
 * !!! 'pvG' is allocated here.
 */
int32 dd_collectVector( Vector **pvG, Vector *v,
			int32 sizeGlobal, int32 *setLocal, int32 nThread )
{
  int32 ret;
  int32 me;
  int32 i, j;
  Vector *vG;
  float64 *bufF;
  int32 *bufI;
  int32 size;

  /*
   * All threads.
   */
  output( "Collect vector." );

  ret = itc_threadRank( &me );
  if (ret) return( ret );

  if (me == 0) {
    /*
     * Master.
     */
    ret = v_createAlloc( pvG, sizeGlobal );
    if (ret) return( ret );
    vG = *pvG;

    bufF = (float64 *) allocMem( vG->nItem * sizeof( float64 ) );
    bufI = (int32 *) allocMem( vG->nItem * sizeof( int32 ) );

    for (i = 0; i < v->nItem; i++) {
      vG->val[setLocal[i]] = v->val[i];
    }

    for (i = 1; i < nThread; i++) {
      ret = itc_recvArrayALNA( bufI, &size, ITC_TypeInt32,
			       i, ITC_TagArray );
      if (ret) return( ret );
      ret = itc_recvArrayALNA( bufF, &size, ITC_TypeFloat64,
			       i, ITC_TagArray );
      if (ret) return( ret );
      for (j = 0; j < size; j++) {
	vG->val[bufI[j]] = bufF[j];
      }
    }

    freeMem( bufI );
    freeMem( bufF );
  } else {
    /*
     * Slaves.
     */
    ret = itc_sendArray( setLocal, v->nItem, ITC_TypeInt32,
			 0, ITC_TagArray );
    if (ret) return( ret );
    ret = itc_sendArray( v->val, v->nItem, ITC_TypeFloat64,
			 0, ITC_TagArray );
    if (ret) return( ret );
  }

  return( 0 );
}
