#include "itc.h"
#include "matrix_csr.h"
#include "memutils.h"
#include "errutils.h"

/*#include <mpi.h>*/

/*********************************************************************
 * Inter-thread communication routines using MPI.
 *********************************************************************/

int32 itc_init( int32 *pargc, char ***pargv )
{
  int32 ret;

  ret = MPI_Init( pargc, pargv );
  if (ret) g_error = ErrorITC;

  return( ret );
}

int32 itc_finish()
{
  int32 ret;

  ret = MPI_Finalize();
  if (ret) g_error = ErrorITC;

  return( ret );
}

int32 itc_abort( int32 r )
{
  int32 ret;

  ret = MPI_Abort( MPI_COMM_WORLD, r );

  return( ret );
}

int32 itc_numberOfThreads( int32 *n )
{
  int32 ret;

  ret = MPI_Comm_size( MPI_COMM_WORLD, n );
  if (ret) g_error = ErrorITC;

  return( ret );
}

int32 itc_threadRank( int32 *rank )
{
  int32 ret;

  ret = MPI_Comm_rank( MPI_COMM_WORLD, rank );
  if (ret) g_error = ErrorITC;

  return( ret );
}

int32 itc_barrier()
{
  int32 ret;

  ret = MPI_Barrier( MPI_COMM_WORLD );

  return( ret );
}

int32 itc_sizeOfITCtype( int32 *size, int32 itct )
{
  int32 ret = 0;

  switch (itct) {
  case ITC_TypeInt32:
    *size = sizeof( int32 );
    break;
  case ITC_TypeFloat64:
    *size = sizeof( float64 );
    break;
  default:
    *size = sizeof( float64 );
    g_error = ErrorITC;
    ret = -1;
  }

  return( ret );
}

int32 itc_ITCtype2MPItype( MPI_Datatype *mpit, int32 itct )
{
  int32 ret = 0;

  switch (itct) {
  case ITC_TypeInt32:
    *mpit = MPI_INT;
    break;
  case ITC_TypeFloat64:
    *mpit = MPI_DOUBLE;
    break;
  default:
    *mpit = MPI_BYTE;
    g_error = ErrorITC;
    ret = -1;
  }

  return( ret );
}

int32 itc_ITCop2MPIop( MPI_Op *mpio, int32 itco )
{
  int32 ret = 0;

  switch (itco) {
  case ITC_ReduceSum:
    *mpio = MPI_SUM;
    break;
  default:
    *mpio = MPI_SUM;
    g_error = ErrorITC;
    ret = -1;
  }

  return( ret );
}

int32 itc_reduce( void *operand, void *result, int32 count, int32 type,
		  int32 op, int32 root )
{
  int32 ret = 0;
  MPI_Datatype mpiType;
  MPI_Op mpiOp;

  ret = itc_ITCtype2MPItype( &mpiType, type );
  if (ret) goto error;

  ret = itc_ITCop2MPIop( &mpiOp, op );
  if (ret) goto error;

  if (root == ITC_ReduceRootAll) {
    ret = MPI_Allreduce( operand, result, count, mpiType, mpiOp,
			 MPI_COMM_WORLD );
  } else {
    ret = MPI_Reduce( operand, result, count, mpiType, mpiOp, root,
		      MPI_COMM_WORLD );
  }
  return( ret );
 error:
  g_error = ErrorITC;
  return( ret );
}

int32 itc_mCSR_sendMatrix( Matrix *m, int32 rank, int32 tag )
{
  int32 ret = 0;
  int32 buf[4]; /* !!! */
  StorageCSR *data = (StorageCSR *) m->data;
  
  buf[0] = m->nRow;
  buf[1] = m->nCol;
  buf[2] = m->nItem;
  buf[3] = m->nAlloc;

#ifdef DEBUG_ITC
  output( "matrix: to: %d tag: %d %d %d %d %d",
	  rank, tag, buf[0], buf[1], buf[2], buf[3] );
#endif
  
  ret = itc_sendArray( buf, 4, ITC_TypeInt32, rank, tag );
  if (ret) goto error;

  ret = itc_sendArray( data->row, m->nRow + 1, ITC_TypeInt32,
		       rank, tag + 2 );
  if (ret) goto error;
  ret = itc_sendArray( data->col, m->nAlloc, ITC_TypeInt32, rank, tag + 4 );
  if (ret) goto error;
  ret = itc_sendArray( data->nInRow, m->nRow + 1, ITC_TypeInt32,
		       rank, tag + 6 );
  if (ret) goto error;
  ret = itc_sendArray( data->val, m->nAlloc, ITC_TypeFloat64, rank, tag + 8 );
  if (ret) goto error;

  return( ret );
 error:
  g_error = ErrorITC;
  return( ret );
}

int32 itc_mCSR_recvMatrix( Matrix **pm, int32 rank, int32 tag )
{
  int32 ret = 0;
  int32 size;
  int32 buf[4]; /* !!! */
  Matrix *m;
  StorageCSR *data;

  ret = itc_recvArrayNA( buf, &size, 4, ITC_TypeInt32, rank, tag );
  if (ret) goto error;

#ifdef DEBUG_ITC
  output( "matrix: from: %d tag: %d %d %d %d %d",
	  rank, tag, buf[0], buf[1], buf[2], buf[3] );
#endif

  ret = mCSR_createAlloc( pm, buf[0], buf[1], buf[3] );
  if (ret) goto error;
  m = *pm;
  m->nItem = buf[2];
  m->type = TypeMatrixCSR;
  data = (StorageCSR *) m->data;

  ret = itc_recvArrayNA( data->row, &size, m->nRow + 1, ITC_TypeInt32,
			 rank, tag + 2 );
  if (ret) goto error;
  ret = itc_recvArrayNA( data->col, &size, m->nAlloc, ITC_TypeInt32,
			 rank, tag + 4 );
  if (ret) goto error;
  ret = itc_recvArrayNA( data->nInRow, &size, m->nRow + 1, ITC_TypeInt32,
			 rank, tag + 6 );
  if (ret) goto error;
  ret = itc_recvArrayNA( data->val, &size, m->nAlloc, ITC_TypeFloat64,
			 rank, tag + 8 );
  if (ret) goto error;

  return( ret );
 error:
  g_error = ErrorITC;
  return( ret );
}

int32 itc_sendValue( void *a, int32 type, int32 rank, int32 tag )
{
  int32 ret = 0;
  MPI_Datatype mpiType;

  ret = itc_ITCtype2MPItype( &mpiType, type );
  if (ret) goto error;

#ifdef DEBUG_ITC
  output( "value: to: %d tag: %d",
	  rank, tag);
#endif

  ret = MPI_Send( a, 1, mpiType, rank, tag, MPI_COMM_WORLD );

  return( ret );
 error:
  g_error = ErrorITC;
  return( ret );
}

int32 itc_recvValue( void *a, int32 type,  int32 rank, int32 tag )
{
  int32 ret;
  MPI_Datatype mpiType;
  MPI_Status status;

  ret = itc_ITCtype2MPItype( &mpiType, type );
  if (ret) goto error;

#ifdef DEBUG_ITC
  output( "value: from: %d tag: %d",
	  rank, tag);
#endif

  ret = MPI_Recv( a, 1, mpiType, rank, tag, MPI_COMM_WORLD, &status );

  return( ret );
 error:
  g_error = ErrorITC;
  return( ret );
}

int32 itc_sendArray( void *a, int32 length, int32 type,
		     int32 rank, int32 tag )
{
  int32 ret = 0;
  MPI_Datatype mpiType;

  ret = itc_ITCtype2MPItype( &mpiType, type );
  if (ret) goto error;

  ret = itc_sendValue( &length, ITC_TypeInt32, rank, tag );
  if (ret) goto error;

#ifdef DEBUG_ITC
  output( "array: to: %d tag: %d len: %d",
	  rank, tag+1, length );
#endif

  ret = MPI_Send( a, length, mpiType, rank, tag + 1, MPI_COMM_WORLD );

  return( ret );
 error:
  g_error = ErrorITC;
  return( ret );
}

/*
 * !!! 'pa' is allocated here.
 */
int32 itc_recvArray( void **pa, int32 *length, int32 type,
		     int32 rank, int32 tag )
{
  int32 ret;
  MPI_Datatype mpiType;
  MPI_Status status;
  int32 sizeOfType;

  ret = itc_sizeOfITCtype( &sizeOfType, type );
  if (ret) goto error;

  ret = itc_ITCtype2MPItype( &mpiType, type );
  if (ret) goto error;

  ret = itc_recvValue( length, ITC_TypeInt32, rank, tag );
  if (ret) goto error;

#ifdef DEBUG_ITC
  output( "array: from: %d tag: %d len: %d, size: %d",
	  rank, tag+1, *length, sizeOfType );
#endif

  *pa = allocMem( *length * sizeOfType );

  ret = MPI_Recv( *pa, *length, mpiType, rank, tag + 1,
		  MPI_COMM_WORLD, &status );

  return( ret );
 error:
  g_error = ErrorITC;
  return( ret );
}

/*
 * Receive already allocated array, limits are checked.
 */
int32 itc_recvArrayNA( void *a, int32 *length, int32 supposedLength,
		       int32 type, int32 rank, int32 tag )
{
  int32 ret;
  MPI_Datatype mpiType;
  MPI_Status status;

  ret = itc_ITCtype2MPItype( &mpiType, type );
  if (ret) goto error;

  ret = itc_recvValue( length, ITC_TypeInt32, rank, tag );
  if (ret) goto error;

#ifdef DEBUG_ITC
  output( "arrayNA: from: %d tag: %d len: %d, slen: %d",
	  rank, tag+1, *length, supposedLength );
#endif

  if (*length != supposedLength) {
    g_error = ErrorITC;
    ret = -1;
    return( ret );
  }

  ret = MPI_Recv( a, *length, mpiType, rank, tag + 1,
		  MPI_COMM_WORLD, &status );

  return( ret );
 error:
  g_error = ErrorITC;
  return( ret );
}

/*
 * Receive already allocated array, limits are not checked.
 */
int32 itc_recvArrayALNA( void *a, int32 *length, int32 type,
			 int32 rank, int32 tag )
{
  int32 ret;
  MPI_Datatype mpiType;
  MPI_Status status;

  ret = itc_ITCtype2MPItype( &mpiType, type );
  if (ret) goto error;

  ret = itc_recvValue( length, ITC_TypeInt32, rank, tag );
  if (ret) goto error;

#ifdef DEBUG_ITC
  output( "arrayALNA: from: %d tag: %d len: %d",
	  rank, tag+1, *length );
#endif

  ret = MPI_Recv( a, *length, mpiType, rank, tag + 1,
		  MPI_COMM_WORLD, &status );

  return( ret );
 error:
  g_error = ErrorITC;
  return( ret );
}
