#include "errutils.h"
#include "memutils.h"
#include "ioutils.h"
#include "timer.h"
#include "matrix_csr.h"
#include "distdata.h"
#include "solver.h"

#include <stdio.h>

int32 main( int32 argc, char **argv )
{
  int32 ret;
  int32 i;
  char *nameResult = 0, *nameMatrix = 0, *nameRHS = 0;
  Matrix *matrixGlobal = 0;
  Vector *resultGlobal = 0;
  Vector *rhsGlobal = 0;
  Matrix *matrixLocal = 0, *matrixLR = 0;
  Vector *resultLocal = 0;
  Vector *rhsLocal = 0;
  CommList *clI = 0, *clO = 0;
  int32 sizeGlobal, sizeLocal, sizeLR;
  int32 *setLocal = 0, *setLR = 0;
  int32 me, nThread;
  Timer timer;

  /*
   * All threads.
   */
  ret = itc_init( &argc, &argv );
  if (ret) return( ret );

  ret = itc_threadRank( &me );
  if (ret) goto main_end;

  g_threadRankForOutput = me;
  g_stdoutTermOutput = !me;
  g_stderrTermOutput = !me;
  sprintf( g_stdoutLogFileName, "log.%d", me );
  sprintf( g_stderrLogFileName, "log.%d", me );
  remove( g_stdoutLogFileName );
  remove( g_stderrLogFileName );

  ret = itc_numberOfThreads( &nThread );
  if (ret) goto main_end;

  if (nThread <= 0) nThread = 1;

  output( "argc: %d", argc );
  for (i = 0; i < argc; i++) {
    output( "argv: %s", argv[i] );
  }

  /* Ugly hack for broken Score. */
  if (argc > 4) argc = 4;

  if (argc != 4) {
    errput( "Usage: <program name> RESULT MATRIX RHS" );
    ret = -1;
    itc_abort( ret );
  }

  timerStart( &timer );
  output( "Master loads data..." );
  if (me == 0) {
    /*
     * Master.
     */
    output( "I am thread %d - big slave master!", me );
    output( "I have %d slave(s)!", nThread - 1 );
    if (nThread == 1) {
      output( "Well, I am probably not so big..." );
    }
    nameResult = argv[1];
    nameMatrix = argv[2];
    nameRHS = argv[3];
    
    ret = mCSR_load( &matrixGlobal, nameMatrix );
    if (ret) goto main_end;
    output( "Global matrix has %d nonzeros allocated.", matrixGlobal->nAlloc );

    if (nThread > matrixGlobal->nRow) {
      errput( "Too much threads, don't you think?" );
      ret = -1;
      itc_abort( ret );
    }
    
    ret = v_load( &rhsGlobal, nameRHS );
    if (ret) goto main_end;
  }
  /*
   * All threads.
   */
  itc_barrier();
  output( "Data loading done." );
  timerPause( &timer );
  output( "Load time: t: %f u: %f s: %f",
	  timer.ticksEnd / (float64) CLK_TCK,
	  timer.userEnd / (float64) CLK_TCK,
	  timer.systemEnd / (float64) CLK_TCK );
  timerStart( &timer );
  if (nThread > 1) {
    ret = dd_distribMatrix( &matrixLocal, &matrixLR, &clI, &clO,
			    &sizeLocal, &setLocal,
			    &sizeLR, &setLR,
			    &sizeGlobal, matrixGlobal, nThread );
    if (ret) goto main_end;
    
    ret = dd_distribVector( &rhsLocal, rhsGlobal, sizeLocal, setLocal,
			    nThread );
    if (ret) goto main_end;
    
    if (me == 0) {
      /*
       * Master.
       */
      mCSR_freeDestroy( &matrixGlobal );
      v_freeDestroy( &rhsGlobal );
    }
    
    /*
     * All threads.
     */
    ret = dd_translateGlobal2Local( clI, clI, sizeLR, setLR,
				    sizeGlobal );
    if (ret) goto main_end;
    
    ret = dd_translateGlobal2Local( clO, clO, sizeLocal, setLocal,
				    sizeGlobal );
    if (ret) goto main_end;

    ret = v_createAlloc( &resultLocal, rhsLocal->nItem );
    if (ret) goto main_end;

    timerPause( &timer );
    output( "Setup time: t: %f u: %f s: %f",
	    timer.ticksEnd / (float64) CLK_TCK,
	    timer.userEnd / (float64) CLK_TCK,
	    timer.systemEnd / (float64) CLK_TCK );
    output( "Local matrix has %d nonzeros allocated.", matrixLocal->nAlloc );
    output( "Interface matrix has %d nonzeros allocated.", matrixLR->nAlloc );

    timerStart( &timer );
    
    output( "Using distributed PCG." );
    ret = sol_d_PCG( resultLocal, matrixLocal, matrixLR, rhsLocal,
		     0, 0.00000001, 0, clI, clO );
    if (ret) goto main_end;
    timerPause( &timer );
    output( "Solution time: t: %f u: %f s: %f",
	    timer.ticksEnd / (float64) CLK_TCK,
	    timer.userEnd / (float64) CLK_TCK,
	    timer.systemEnd / (float64) CLK_TCK );
    timerStart( &timer );
    
    ret = dd_collectVector( &resultGlobal, resultLocal,
			    sizeGlobal, setLocal, nThread );
    if (ret) goto main_end;

  } else {
    ret = v_createAlloc( &resultGlobal, rhsGlobal->nItem );
    if (ret) goto main_end;

    timerPause( &timer );
    output( "Setup time: t: %f u: %f s: %f",
	    timer.ticksEnd / (float64) CLK_TCK,
	    timer.userEnd / (float64) CLK_TCK,
	    timer.systemEnd / (float64) CLK_TCK );

    timerStart( &timer );

    output( "Using normal PCG." );
    ret = sol_PCG( resultGlobal, matrixGlobal, rhsGlobal,
		     0, 0.00000001, 0 );
    if (ret) goto main_end;
    timerPause( &timer );
    output( "Solution time: t: %f u: %f s: %f",
	    timer.ticksEnd / (float64) CLK_TCK,
	    timer.userEnd / (float64) CLK_TCK,
	    timer.systemEnd / (float64) CLK_TCK );
    timerStart( &timer );

    mCSR_freeDestroy( &matrixGlobal );
    v_freeDestroy( &rhsGlobal );
  }
  timerPause( &timer );
  output( "Collect time: t: %f u: %f s: %f",
	  timer.ticksEnd / (float64) CLK_TCK,
	  timer.userEnd / (float64) CLK_TCK,
	  timer.systemEnd / (float64) CLK_TCK );
  output( "Master saves results..." );
  if (me == 0) {
    /*
     * Master.
     */
    ret = v_save( resultGlobal, nameResult );
    if (ret) goto main_end;

    v_freeDestroy( &resultGlobal );
  }
  /*
   * All threads.
   */
  output( "Result saving done." );
 main_end:
  if (nThread > 1) {
    mCSR_freeDestroy( &matrixLocal );
    mCSR_freeDestroy( &matrixLR );
    v_freeDestroy( &resultLocal );
    v_freeDestroy( &rhsLocal );
    freeMem( setLocal );
    freeMem( setLR );
    for (i = 0; i < clI->nList; i++) {
      freeMem( clI->cs[i].set );
    }
    freeMem( clI->cs );
    freeMem( clI );
    for (i = 0; i < clO->nList; i++) {
      freeMem( clO->cs[i].set );
    }
    freeMem( clO->cs );
    freeMem( clO );
  }
  output( "All memory freed." );
  
  if (ret) pError( "main():" );

  output( "All work done." );

  ret = itc_finish();
  if (ret) pError( "main():" );

  return( ret );
}
