#include "memutils.h"
#include "errutils.h"
#include "matrix_csr.h"

#include <stdio.h>

/*
 * Create empty CSR matrix.
 * 'data' are not allocated.
 */
int32 mCSR_create( Matrix **pm )
{
  if ((*pm = allocMatrix()) == 0) {
    g_error = ErrorAlloc;
    return( -1 );
  }
  (*pm)->type = TypeMatrixCSR;
  (*pm)->nRow = 0;
  (*pm)->nCol = 0;
  (*pm)->nAlloc = 0;
  (*pm)->nItem = 0;
  (*pm)->data = 0;

  return( 0 );
}

/*
 * Allocate space for 'n' items in CSR matrix '*m' (must be allocated).
 * Storage is always allocated (no check).
 */
int32 mCSR_alloc( Matrix *m, int32 nRow, int32 nCol, int32 n )
{
  StorageCSR *data;

  if (!m) {
    g_error = ErrorNullPointer;
    return( -1 );
  }
  if (m->type != TypeMatrixCSR) {
    g_error = ErrorWrongType;
    return( -1 );
  }
  
  m->nRow = nRow;
  m->nCol = nCol;

  m->data = allocMem( sizeof( StorageCSR ) );
  if (m->data == 0) {
    g_error = ErrorAlloc;
    return( -1 );
  }

  data = (StorageCSR *) m->data;

  data->row = (int32 *) allocMem( (nRow + 1) * sizeof( int32 ) );
  if (data->row == 0) {
    g_error = ErrorAlloc;
    return( -1 );
  }
  data->nInRow = (int32 *) allocMem( (nRow + 1) * sizeof( int32 ) );
  if (data->nInRow == 0) {
    g_error = ErrorAlloc;
    return( -1 );
  }
  data->col = (int32 *) allocMem( n * sizeof( int32 ) );
  if (data->col == 0) {
    g_error = ErrorAlloc;
    return( -1 );
  }
  data->val = (float64 *) allocMem( n * sizeof( float64 ) );
  if (data->val == 0) {
    g_error = ErrorAlloc;
    return( -1 );
  }

  m->nAlloc = n;
  m->nItem = 0;

  return( 0 );
}

/*
 * Create and allocate CSR matrix.
 */
int32 mCSR_createAlloc( Matrix **pm, int32 nRow, int32 nCol, int32 n )
{
  int32 ret;

  if ((ret = mCSR_create( pm ))) return( ret );
  if ((ret = mCSR_alloc( *pm, nRow, nCol, n ))) return( ret );

  return( 0 );
}

/*
 * Free memory used by storage of CSR matrix.
 */
int32 mCSR_free( Matrix *m )
{
  StorageCSR *data;

  if (!m) {
    return( 0 );
  }
  if (!m->data) {
    return( 0 );
  }
  if (m->type != TypeMatrixCSR) {
    g_error = ErrorWrongType;
    return( -1 );
  }
  
  data = (StorageCSR *) m->data;
  freeMem( data->row );
  freeMem( data->nInRow );
  freeMem( data->col );
  freeMem( data->val );
  freeMem( m->data );

  m->nRow = 0;
  m->nCol = 0;
  m->nAlloc = 0;
  m->nItem = 0;

  return( 0 );
}

/*
 * Free CSR matrix data.
 */
int32 mCSR_destroy( Matrix **pm )
{
  if (!pm) {
    return( 0 );
  }

  freeMatrix( *pm );
  *pm = 0;

  return( 0 );
}

/*
 * Free and destroy CSR matrix.
 */
int32 mCSR_freeDestroy( Matrix **pm )
{
  int32 ret;

  if ((ret = mCSR_free( *pm ))) return( ret );
  if ((ret = mCSR_destroy( pm ))) return( ret );

  return( 0 );
}

/*
 * Initialize already allocated (no check) 'row' pointers.
 * 'init' .. supposed number of items in rows.
 */
int32 mCSR_initRowPointers( int32 *row, int32 *init, int32 nRow, int32 nItem )
{
  int32 i, j;

  if (nRow <= 0) {
    g_error = ErrorBadLimit;
    return( -1 );
  }
  row[0] = 0;
  j = 0;
  for (i = 1; i <= nRow; i++) {
    j += init[i-1];
    row[i] = j;
  }
  if (row[nRow-1] > nItem) {
    g_error = ErrorRowInit;
    return( -1 );
  }
  row[nRow] = nItem;

  return( 0 );
}

/*
 * Find item at position ('row','col'). 'data' must be allocated (no check).
 * Return: index in 'val' field or -1.
 */
int32 mCSR_find( StorageCSR *data, int32 row, int32 col )
{
  int32 i;
  int32 *pc, r, *pn;

  r = data->row[row];
  pc = &data->col[r];
  pn = data->nInRow;
  i = 0;
  while (i < pn[row]) {
    if (col == pc[i]) return( r + i );
    i++;
  }

  return( -1 );
}

/*
 * Insert new item 'val' to ('row', 'col'). 'm' and its storage must be
 * allocated and of correct type (no check).
 */
int32 mCSR_insert( Matrix *m, int32 row, int32 col, float64 val )
{
  int32 ret = 0;
  int32 i;
  StorageCSR *data = (StorageCSR *) m->data;
  
  if ((row >= m->nRow) || (col >= m->nCol)) {
    g_error = ErrorBadLimit;
    return( -1 );
  }

  i = mCSR_find( data, row, col );

  if (i >= 0) {
    data->val[i] = val;
  } else {
    ret = mCSR_insertNew( m, row, col, val );
  }

  return( ret );
}

/*
 * Add item 'val' to ('row', 'col'). 'm' and its storage must be
 * allocated and of correct type (no check).
 */
int32 mCSR_addToPlace( Matrix *m, int32 row, int32 col, float64 val )
{
  int32 ret = 0;
  int32 i;
  StorageCSR *data = (StorageCSR *) m->data;
  
  if ((row >= m->nRow) || (col >= m->nCol)) {
    g_error = ErrorBadLimit;
    return( -1 );
  }

  i = mCSR_find( data, row, col );

  if (i >= 0) {
    data->val[i] += val;
  } else {
    ret = mCSR_insertNew( m, row, col, val );
  }

  return( ret );
}

/*
 * Insert new item. Make space if required. No checks.
 */
int32 mCSR_insertNew( Matrix *m, int32 row, int32 col, float64 val )
{
  StorageCSR *data = (StorageCSR *) m->data;
  int32 ret;
  int32 i, *pr, *pc, *pn;
  float64 *pv;

  pr = data->row;
  pn = data->nInRow;
  pc = data->col;
  pv = data->val;

  if ((pn[row] + pr[row]) >= pr[row+1]) {
    ret = mCSR_makeSpace( row, m->nRow, &m->nAlloc, 100, data );
    if (!ret) return( ret );
    pc = data->col;
    pv = data->val;
  }

  i = pr[row]+pn[row];
  pc[i] = col;
  pv[i] = val;
  pn[row]++;
  m->nItem++;

  return( 0 );
}

/*
 * Make space for more items. No checks.
 */
int32 mCSR_makeSpace( int32 start, int32 nRow, int32 *nAlloc, int32 shift,
		      StorageCSR *data )
{
  int32 i;
  int32 *pr, *pc;
  float64 *pv;

  data->col = (int32 *) reallocMem( data->col,
				    (*nAlloc + shift) * sizeof( int32 ) );
  if (!data->col) {
    g_error = ErrorAlloc;
    return( -1 );
  }
  data->val = (float64 *) reallocMem( data->val,
				      (*nAlloc + shift) * sizeof( float64 ) );
  if (!data->val) {
    g_error = ErrorAlloc;
    return( -1 );
  }

  pv = data->val;
  pr = data->row;
  pc = data->col;

  for (i = (start+1); i <= nRow; i++) {
    pr[i] += shift;
  }
  *nAlloc += shift;
  for (i = pr[start+1]; i < *nAlloc; i++) {
    pc[i] = pc[i-shift];
    pv[i] = pv[i-shift];
  }

  return( 0 );
}

/*
 * Save CSR matrix to a file.
 */
int32 mCSR_save( Matrix *m, const char *fileName )
{
  StorageCSR *data;
  float64 *pv;
  int32 *pr, *pn, *pc;
  FILE *file;
  int32 i, j;

  if (!m) {
    g_error = ErrorNullPointer;
    return( -1 );
  }
  if (!m->data) {
    g_error = ErrorNullPointer;
    return( -1 );
  }
  if (m->type != TypeMatrixCSR) {
    g_error = ErrorWrongType;
    return( -1 );
  }

  data = (StorageCSR *) m->data;

  if ((file = fopen( fileName, "w" )) == 0) {
    g_error = ErrorFileOpen;
    return( -1 );
  }

  pv = data->val;
  pr = data->row;
  pn = data->nInRow;
  pc = data->col;

  fprintf( file, "%d %d\n", m->nRow, m->nCol );
  fprintf( file, "%d\n", m->nItem );
  for (i = 0; i < m->nRow; i++) {
    for (j = pr[i]; j < (pr[i] + pn[i]); j++) {
      fprintf( file, "%d %d %.12e\n", i, pc[j], pv[j] );
    }
  }
  fclose( file );

  return( 0 );
}

/*
 * Load CSR matrix from a file. Matrix is created and allocated here.
 */
int32 mCSR_load( Matrix **pm, const char *fileName )
{
  MItem *sortList;
  Matrix *m;
  StorageCSR *data;
  int32 k;  
  float64 *pv;
  int32 *pr, *pn, *pc;
  FILE *file;
  int32 i, j;
  int32 ret;
  
  if ((file = fopen( fileName, "r" )) == 0) {
    g_error = ErrorFileOpen;
    return( -1 );
  }

  ret = mCSR_create( pm );
  if (ret) return( ret );

  m = *pm;
  fscanf( file, "%d %d", &m->nRow, &m->nCol );
  fscanf( file, "%d", &m->nAlloc );

  if ((m->nRow <= 0) || (m->nCol <= 0) || (m->nAlloc <= 0)) {
    g_error = ErrorBadLimit;
    fclose( file );
    return( -1 );
  }

  ret = mCSR_alloc( m, m->nRow, m->nCol, m->nAlloc );
  if (ret) return( ret );

  data = (StorageCSR *) m->data;
  pv = data->val;
  pr = data->row;
  pn = data->nInRow;
  pc = data->col;

  /*
   * Data will be read in 'sortList' and sorted w.r.t rows and then columns
   * in a row.
   */
  sortList = (MItem *) allocMem( m->nAlloc * sizeof( MItem ) );
  for (i = 0; i < m->nAlloc; i++) {
    fscanf( file, "%d %d" SCANF64,
	    &sortList[i].row, &sortList[i].col, &sortList[i].val );
  }
  qsort( sortList, m->nAlloc, sizeof( MItem ), &compareMItem );
/*    for (i = 0; i < m->nAlloc; i++) { */
/*      printf( "%d %d %f\n", */
/*  	    sortList[i].row, sortList[i].col, sortList[i].val ); */
/*    } */
  if (g_error == ErrorCompare) {
    fclose( file );
    return( -1 );
  }

  /*
   * Transform it to CSR format.
   */
  k = 0;
  pr[0] = 0;
/*    printf( "%d %d %d\n", m->nRow, m->nCol, m->nAlloc ); */
  for (i = 0; i < m->nRow; i++) {
    pn[i] = 0;
    if (k >= m->nAlloc) break;
    j = k;
    while (sortList[k].row == i) {
      pv[k] = sortList[k].val;
      pc[k] = sortList[k].col;
      k++;
      if (k >= m->nAlloc) break;
    }
    pn[i] = k - j;
    pr[i+1] = k;
/*      printf( "%d %d\n", pn[i], pr[i] ); */
  }
  m->nItem = pr[m->nRow];
 
  fclose( file );

  freeMem( sortList );

  if (m->nItem != m->nAlloc) {
    g_error = ErrorDataRead;
    return( -1 );
  }

  return( 0 );
}

/*
 * Return in 'n' a number of nonzeros in column 'col' of matrix 'm'.
 */
/*  int32 mCSR_nInCol( int32 *n, Matrix *m, int32 col ) */
/*  { */
/*    return( 0 ); */
/*  } */

/*  int32 findInList( int32 n, int32 *list, int32 nList ) */
/*  { */
/*    int32 i; */

/*    i = 0; */
/*    while (i < nList) { */
/*      if (list[i] == ) break; */
/*      i++; */
/*    } */
/*    return( i ); */
/*  } */

/*
 * Create CSR submatrix of CSR matrix.
 * O:    pm .. resulting submatrix
 * I:     m .. matrix
 *    nSRow .. number of selected rows
 *    nSCol .. number of selected columns.
 *       sr .. row indices
 *       sc .. col indices of submatrix in matrix
 * !!! 'pm' is allocated here.
 */
int32 mCSR_submatrix( Matrix **pm, Matrix *m, int32 nSRow, int32 nSCol,
		      int32 *sr, int32 *sc, int32 colSelType )
{
  Matrix *m2;
  StorageCSR *data, *data2;
  float64 *pv;
  int32 *pr, *pn, *pc;
  float64 *pv2;
  int32 *pr2, *pn2, *pc2;
  int32 r;
  int32 i, j, k, l;
  int32 n;
  int32 *colSel;
  int32 ret;

  /*
   * Check submatrix indices (mutual diff. not checked!).
   */
  for (i = 0; i < nSRow; i++) {
    if ((sr[i] < 0) || (sr[i] >= m->nRow)) {
      g_error = ErrorBadLimit;
      return( -1 );
    }
  }

  if (!colSelType) {
    for (i = 0; i < nSCol; i++) {
      if ((sc[i] < 0) || (sc[i] >= m->nCol)) {
	g_error = ErrorBadLimit;
	return( -1 );
      }
    }
    colSel = (int32 *) allocMem( (m->nCol) * sizeof( int32 ) );
    if (colSel == 0) {
      g_error = ErrorAlloc;
      return( -1 );
    }
    for (i = 0; i < m->nCol; i++) {
      colSel[i] = -1;
    }
    for (i = 0; i < nSCol; i++) {
      colSel[sc[i]] = i;
    }
  } else colSel = sc;

  data = (StorageCSR *) m->data;
  pr = data->row;
  pn = data->nInRow;
  
  /*
   * Number of nonzeros in submatrix.
   */
  n = 0;
  for (i = 0; i < nSRow; i++) {
    r = sr[i];
    pc = &data->col[pr[r]];
    for (j = 0; j < pn[r]; j++) {
      if (colSel[pc[j]] >= 0) n++;
    }
  }

/*    printf( "%d\n", n ); */

  ret = mCSR_createAlloc( pm, nSRow, nSCol, n );
  if (ret) return( ret );

  m2 = *pm;
  data2 = (StorageCSR *) m2->data;
  pr2 = data2->row;
  pn2 = data2->nInRow;
  pc2 = data2->col;
  pv2 = data2->val;

  /*
   * Fill submatrix.
   */
  k = 0;
  pr2[0] = 0;
  for (i = 0; i < nSRow; i++) {
    r = sr[i];
    pc = &data->col[pr[r]];
    pv = &data->val[pr[r]];
    /* Search selected row for selected columns. */
    l = k;
    for (j = 0; j < pn[r]; j++) {
/*        printf( "%d %d\n", pc[j], colSel[pc[j]] ); */
      if (colSel[pc[j]] >= 0) {
	pc2[k] = colSel[pc[j]];
	pv2[k] = pv[j];
	k++;
      }
    }
    pn2[i] = k - l;
    pr2[i+1] = k;
/*      printf( "%d %d %d\n", r, pn2[i], pr2[i] ); */
  }
  m2->nItem = n;

  if (!colSelType) freeMem( colSel );

  return( 0 );
}
