#ifndef _MATRIX_CSR_H_
#define _MATRIX_CSR_H_

#include "matrix.h"

typedef struct {
  int32 *row;
  int32 *col;
  int32 *nInRow;
  float64 *val;
} StorageCSR;

/*
 * Create empty CSR matrix.
 * 'data' are not allocated.
 */
int32 mCSR_create( Matrix **pm );
/*
 * Allocate space for 'n' items in CSR matrix '*m' (must be allocated).
 * Storage is always allocated (no check).
 */
int32 mCSR_alloc( Matrix *m, int32 nRow, int32 nCol, int32 n );
/*
 * Create and allocate CSR matrix.
 */
int32 mCSR_createAlloc( Matrix **pm, int32 nRow, int32 nCol, int32 n );
/*
 * Free memory used by storage of CSR matrix.
 */
int32 mCSR_free( Matrix *m );
/*
 * Free CSR matrix data.
 */
int32 mCSR_destroy( Matrix **pm );
/*
 * Free and destroy CSR matrix.
 */
int32 mCSR_freeDestroy( Matrix **pm );
/*
 * Initialize already allocated (no check) 'row' pointers.
 * 'init' .. supposed number of items in rows.
 */
int32 mCSR_initRowPointers( int32 *row, int32 *init, int32 nRow, int32 nItem );
/*
 * Find item at position ('row','col'). 'data' must be allocated (no check).
 * Return: index in 'val' field or -1.
 */
int32 mCSR_find( StorageCSR *data, int32 row, int32 col );
/*
 * Insert new item 'val' to ('row', 'col'). 'm' and its storage must be
 * allocated and of correct type (no check).
 */
int32 mCSR_insert( Matrix *m, int32 row, int32 col, float64 val );
/*
 * Add item 'val' to ('row', 'col'). 'm' and its storage must be
 * allocated and of correct type (no check).
 */
int32 mCSR_addToPlace( Matrix *m, int32 row, int32 col, float64 val );
/*
 * Insert new item. Make space if required. No checks.
 */
int32 mCSR_insertNew( Matrix *m, int32 row, int32 col, float64 val );
/*
 * Make space for more items. No checks.
 */
int32 mCSR_makeSpace( int32 start, int32 nRow, int32 *nAlloc, int32 shift,
		      StorageCSR *data );
/*
 * Save CSR matrix to a file.
 */
int32 mCSR_save( Matrix *m, const char *fileName );
/*
 * Load CSR matrix from a file. Matrix is created and allocated here.
 */
int32 mCSR_load( Matrix **pm, const char *fileName );
/*
 * Create CSR submatrix of CSR matrix.
 */
int32 mCSR_submatrix( Matrix **pm, Matrix *m, int32 nSRow, int32 nSCol,
		      int32 *sr, int32 *sc, int32 colSelType );

#endif
