#include "ops_la.h"
#include "errutils.h"

#include <string.h>

/*
 * Linear algebra operations. All relevant objects must be allocated
 * and properly initialized (no check).
 */

/*
 * Test sizes for matrix - vector multiplying.
 */
int32 la_testSizes_mul_m_v( int32 nRow, int32 nCol, int32 n1, int32 n2 )
{
  if ((n1 == nRow) && (n2 == nCol))
    return( 0 );
  else
    return( -1 );
}

/*
 * Test sizes for vector - vector sum.
 */
int32 la_testSizes_add_v_v( int32 n1, int32 n2, int32 n3 )
{
  if ((n1 == n2) && (n1 == n3))
    return( 0 );
  else
    return( -1 );
}

/*
 * v_1 = M v_2. Sizes are checked.
 */
int32 la_mul_m_v( Vector *v1, Matrix *m, Vector *v2 )
{
  int32 ret = 0;

  if (la_testSizes_mul_m_v( m->nRow, m->nCol, v1->nItem, v2->nItem )) {
    g_error = ErrorBadMatch;
    return( -1 );
  }

  switch (m->type) {
  case TypeMatrixCSR:
    ret = la_mul_CSR_m_v( v1, m, v2 );
    break;
  default:
    g_error = ErrorWrongType;
    ret = -1;
  }

  return( ret );
}

/*
 * v_1 = M^T v_2. Sizes are checked.
 */
int32 la_mul_mT_v( Vector *v1, Matrix *m, Vector *v2 )
{
  int32 ret = 0;

  if (la_testSizes_mul_m_v( m->nCol, m->nRow, v1->nItem, v2->nItem )) {
    g_error = ErrorBadMatch;
    return( -1 );
  }

  switch (m->type) {
  case TypeMatrixCSR:
    ret = la_mul_CSR_mT_v( v1, m, v2 );
    break;
  default:
    g_error = ErrorWrongType;
    ret = -1;
  }

  return( ret );
}

/*
 * v_1 = M v_2. M has CSR storage (no check).
 */
int32 la_mul_CSR_m_v( Vector *v1, Matrix *m, Vector *v2 )
{
  StorageCSR *data;
  int32 i, j;
  int32 *pc;
  float64 *pv;

  data = (StorageCSR *) m->data;
  
  for (i = 0; i < m->nRow; i++) {
    pc = &data->col[data->row[i]];
    pv = &data->val[data->row[i]];
    v1->val[i] = 0.0;
    for (j = 0; j < data->nInRow[i]; j++) {
      v1->val[i] += pv[j] * v2->val[pc[j]];
    }
  }

  return( 0 );
}

/*
 * v_1 = M^T v_2. M has CSR storage (no check).
 */
int32 la_mul_CSR_mT_v( Vector *v1, Matrix *m, Vector *v2 )
{
  StorageCSR *data;
  int32 i, j;
  int32 *pc;
  float64 *pv;

  data = (StorageCSR *) m->data;

  memset( v1->val, 0, v1->nItem * sizeof( float64 ) );
  for (i = 0; i < m->nRow; i++) {
    pc = &data->col[data->row[i]];
    pv = &data->val[data->row[i]];
    for (j = 0; j < data->nInRow[i]; j++) {
      v1->val[pc[j]] += pv[j] * v2->val[i];
    }
  }

  return( 0 );
}

/*
 * v_1 = v_2 + v_3. Sizes are checked.
 */
int32 la_add_v_v( Vector *v1, Vector *v2, Vector *v3 )
{
  int32 i;

  if (la_testSizes_add_v_v( v1->nItem, v2->nItem, v3->nItem )) {
    g_error = ErrorBadMatch;
    return( -1 );
  }
  
  for (i = 0; i < v1->nItem; i++) {
    v1->val[i] = v2->val[i] + v3->val[i];
  }

  return( 0 );
}

/*
 * v_1 = v_2 - v_3. Sizes are checked.
 */
int32 la_sub_v_v( Vector *v1, Vector *v2, Vector *v3 )
{
  int32 i;

  if (la_testSizes_add_v_v( v1->nItem, v2->nItem, v3->nItem )) {
    g_error = ErrorBadMatch;
    return( -1 );
  }
  
  for (i = 0; i < v1->nItem; i++) {
    v1->val[i] = v2->val[i] - v3->val[i];
  }

  return( 0 );
}

/*
 * v_1 = a v_2 + v_3. Sizes are checked.
 */
int32 la_mul_add_f_v_v( Vector *v1, float64 a, Vector *v2, Vector *v3 )
{
  int32 i;

  if (la_testSizes_add_v_v( v1->nItem, v2->nItem, v3->nItem )) {
    g_error = ErrorBadMatch;
    return( -1 );
  }
  
  for (i = 0; i < v1->nItem; i++) {
    v1->val[i] = a * v2->val[i] + v3->val[i];
  }

  return( 0 );
}

/*
 * dot = (v_1, v_2). Sizes are checked.
 */
int32 la_dot_v_v( float64 *dot, Vector *v1, Vector *v2 )
{
  int32 i;

  if (v1->nItem != v2->nItem) {
    g_error = ErrorBadMatch;
    return( -1 );
  }
  
  *dot = 0.0;
  for (i = 0; i < v1->nItem; i++) {
    *dot += v1->val[i] * v2->val[i];
  }
  
  return( 0 );
}

/*
 * dot = (mask v_1, v_2). Sizes are checked.
 */
int32 la_mask_dot_v_v( float64 *dot, int32 *mask, Vector *v1, Vector *v2 )
{
  int32 i;

  if (v1->nItem != v2->nItem) {
    g_error = ErrorBadMatch;
    return( -1 );
  }
  
  *dot = 0.0;
  for (i = 0; i < v1->nItem; i++) {
    if (mask[i]) *dot += v1->val[i] * v2->val[i];
  }
  
  return( 0 );
}

/*
 * Fill vector with a constant.
 */
int32 la_fill_v( Vector *v, float64 val )
{
  int32 i;

  for (i = 0; i < v->nItem; i++) {
    v->val[i] = val;
  }
  
  return( 0 );
}

/*
 * v_1 = v_2. Sizes are checked.
 */
int32 la_copy_v( Vector *v1, Vector *v2 )
{
  if (v1->nItem != v2->nItem) {
    g_error = ErrorBadMatch;
    return( -1 );
  }

  memcpy( v1->val, v2->val, v1->nItem * sizeof( float64 ) );

  return( 0 );
}
