#include "vector.h"
#include "memutils.h"
#include "errutils.h"

#include <stdio.h>

/*
 * Allocate vector structure.
 */
Vector *allocVector()
{
  Vector *v;
  
  v = (Vector *) allocMem( sizeof( Vector ) );
  v->val = 0;

  return( v );
}

/*
 * Free vector structure.
 */
void freeVector( Vector *v )
{
  freeMem( v );
}

/*
 * Create empty vector.
 */
int32 v_create( Vector **pv )
{
  if ((*pv = allocVector()) == 0) {
    g_error = ErrorAlloc;
    return( -1 );
  }

  (*pv)->nAlloc = 0;
  (*pv)->nItem = 0;

  return( 0 );
}

/*
 * Allocate space for 'n' items in vector 'v' (must be allocated).
 * Sets no. of items to 'n'.
 */
int32 v_alloc( Vector *v, int32 n )
{
  if (!v) {
    g_error = ErrorNullPointer;
    return( -1 );
  }

  v->val = (float64 *) allocMem( n * sizeof( float64 ) );
  if (v->val == 0) {
    g_error = ErrorAlloc;
    return( -1 );
  }

  v->nAlloc = n;
  v->nItem = n;

  return( 0 );
}

/*
 * Create and allocate vector.
 */
int32 v_createAlloc( Vector **pv, int32 n )
{
  int32 ret;

  if ((ret = v_create( pv ))) return( ret );
  if ((ret = v_alloc( *pv, n ))) return( ret );
  
  return( 0 );
}

/*
 * Free vector data.
 */
int32 v_free( Vector *v )
{
  if (!v) {
    return( 0 );
  }
  if (!v->val) {
    return( 0 );
  }
  
  freeMem( v->val );

  v->nAlloc = 0;
  v->nItem = 0;

  return( 0 );
}

/*
 * Free vector structure.
 */
int32 v_destroy( Vector **pv )
{
  if (!pv) {
    return( 0 );
  }

  freeVector( *pv );
  *pv = 0;

  return( 0 );
}

/*
 * Free and destroy vector.
 */
int32 v_freeDestroy( Vector **pv )
{
  int32 ret;

  if ((ret = v_free( *pv ))) return( ret );
  if ((ret = v_destroy( pv ))) return( ret );

  return( 0 );
}

/*
 * Save vector to a file.
 */
int32 v_save( Vector *v, const char *fileName )
{
  FILE *file;
  int32 i;

  if (!v) {
    g_error = ErrorNullPointer;
    return( -1 );
  }
  if (!v->val) {
    g_error = ErrorNullPointer;
    return( -1 );
  }

  if ((file = fopen( fileName, "w" )) == 0) {
    g_error = ErrorFileOpen;
    return( -1 );
  }

  fprintf( file, "%d\n", v->nItem );
  for (i = 0; i < v->nItem; i++) {
    fprintf( file, "%d %.12f\n", i, v->val[i] );
  }
  fclose( file );

  return( 0 );
}

/*
 * Load vector from a file. Vector is created and allocated here.
 * Value indices must be correct (no check of duplicity).
 */
int32 v_load( Vector **pv, const char *fileName )
{
  Vector *v;
  FILE *file;
  int32 i, j;
  float64 val;

  if ((file = fopen( fileName, "r" )) == 0) {
    g_error = ErrorFileOpen;
    return( -1 );
  }

  v_create( pv );
  v = *pv;

  fscanf( file, "%d", &v->nAlloc );
  if (v->nAlloc <= 0) {
    g_error = ErrorBadLimit;
    fclose( file );
    return( -1 );
  }
  
  v_alloc( v, v->nAlloc );

  for (i = 0; i < v->nAlloc; i++) {
    fscanf( file, "%d" SCANF64, &j, &val );
    if ((j < 0) || (j >= v->nAlloc)) {
      fclose( file );
      g_error = ErrorDataRead;
      return( -1 );
    }
    v->val[j] = val;
  }
  v->nItem = v->nAlloc;

  fclose( file );

  return( 0 );
}
